__all__ = [
    "MechaError",
    "InvalidEscapeSequence",
    "UnrecognizedParser",
]


class MechaError(Exception):
    """Base class for all mecha errors."""


class InvalidEscapeSequence(MechaError):
    """Raised when a QuotedStringHandler encounters an invalid escape sequence."""

    characters: str
    index: int

    def __init__(self, characters: str, index: int):
        super().__init__(characters, index)
        self.characters = characters
        self.index = index


class UnrecognizedParser(MechaError):
    """Raised when delegating to an unrecognized parser."""

    parser: str

    def __init__(self, parser: str):
        super().__init__(parser)
        self.parser = parser



__all__ = [
    "Diagnostic",
    "DiagnosticCollection",
    "DiagnosticLabelContainer",
    "DiagnosticError",
]


from dataclasses import dataclass, field, replace
from types import TracebackType
from typing import Any, Iterator, List, Literal, Optional, Type

from beet import Container, FormattedPipelineException
from tokenstream import UNKNOWN_LOCATION, SourceLocation


class DiagnosticLabelContainer(Container[Any, str]):
    """Container for diagnostic labels."""

    def normalize_key(self, key: Any) -> Any:
        if isinstance(key, SourceLocation):
            return key, key

        location = getattr(key, "location", None)
        end_location = getattr(key, "end_location", None)

        if location and end_location:
            return location, end_location

        return key


@dataclass
class Diagnostic(Exception):
    """Exception that can be raised to report messages."""

    level: Literal["info", "warn", "error"]
    message: str

    rule: Optional[str] = None
    hint: Optional[str] = None
    filename: Optional[str] = None
    location: SourceLocation = UNKNOWN_LOCATION
    end_location: SourceLocation = UNKNOWN_LOCATION
    labels: DiagnosticLabelContainer = field(default_factory=DiagnosticLabelContainer)

    def __str__(self) -> str:
        return self.format_message()

    def format_message(self) -> str:
        """Return the formatted message."""
        message = self.message
        if self.rule:
            message += f" ({self.rule})"
        return message

    def format_location(self) -> str:
        """Return the formatted location of the reported message."""
        if self.filename:
            location = self.filename
            if not self.location.unknown:
                location += f":{self.location.lineno}:{self.location.colno}"
        elif not self.location.unknown:
            location = f"line {self.location.lineno}, column {self.location.colno}"
            if self.hint:
                location = f'File "{self.hint}", {location}'
        elif self.hint:
            location = self.hint
        else:
            location = ""

        return location

    def with_defaults(
        self,
        rule: Optional[str] = None,
        hint: Optional[str] = None,
        filename: Optional[str] = None,
    ) -> "Diagnostic":
        """Set default values for unspecified attributes."""
        return replace(
            self,
            rule=self.rule or rule,
            hint=self.hint or hint,
            filename=self.filename or filename,
        )


@dataclass
class DiagnosticCollection(Exception):
    """Exception that can be raised to group multiple diagnostics."""

    exceptions: List[Diagnostic] = field(default_factory=list)

    rule: Optional[str] = None
    hint: Optional[str] = None
    filename: Optional[str] = None

    def add(self, exc: Diagnostic) -> Diagnostic:
        """Add diagnostic."""
        exc = exc.with_defaults(rule=self.rule, hint=self.hint, filename=self.filename)
        self.exceptions.append(exc)
        return exc

    def extend(self, other: "DiagnosticCollection"):
        """Combine diagnostics from another collection."""
        self.exceptions.extend(other.exceptions)

    def get_all_errors(self) -> Iterator[Diagnostic]:
        """Yield all the diagnostics with a severity level of "error"."""
        for exc in self.exceptions:
            if exc.level == "error":
                yield exc

    def __enter__(self) -> "DiagnosticCollection":
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ):
        if not exc_type:
            if self.exceptions:
                raise self

    def __str__(self) -> str:
        term = (
            "error"
            if all(exc.level == "error" for exc in self.exceptions)
            else "diagnostic"
        )
        term += "s" * (len(self.exceptions) > 1)
        return f"Reported {len(self.exceptions)} {term}."


class DiagnosticError(FormattedPipelineException):
    """Wrap a collection of error diagnostics to abort the build."""

    diagnostics: DiagnosticCollection

    def __init__(self, diagnostics: DiagnosticCollection, show_details: bool = True):
        super().__init__(diagnostics)
        self.diagnostics = diagnostics

        if show_details:
            details = "\n".join(
                f"{diagnostic.format_location()}: {diagnostic.format_message()}"
                for diagnostic in diagnostics.exceptions
            )
            self.message = f"{diagnostics}\n\n{details}"
        else:
            self.message = str(diagnostics)

        self.format_cause = True



__all__ = [
    "beet_default",
]


import logging

from beet import Context

from .api import Mecha
from .diagnostic import DiagnosticCollection, DiagnosticError

logger = logging.getLogger("mecha")


def beet_default(ctx: Context):
    mc = ctx.inject(Mecha)

    mc.compile(ctx.data, report=mc.diagnostics)

    yield

    for diagnostic in mc.diagnostics.exceptions:
        args = {"message": diagnostic.format_message()}
        extra = {"annotate": diagnostic.format_location()}

        if diagnostic.level == "error":
            logger.error("%(message)s", args, extra=extra)
        elif diagnostic.level == "warn":
            logger.warn("%(message)s", args, extra=extra)
        elif diagnostic.level == "info":
            logger.info("%(message)s", args, extra=extra)

    if errors := list(mc.diagnostics.get_all_errors()):
        raise DiagnosticError(DiagnosticCollection(errors), show_details=False)



__all__ = [
    "Mecha",
    "MechaOptions",
]


import logging
import os
import pickle
from contextlib import contextmanager
from dataclasses import InitVar, dataclass
from pathlib import Path
from typing import (
    Any,
    Dict,
    Iterator,
    List,
    Literal,
    Optional,
    Type,
    TypeVar,
    Union,
    overload,
)

from beet import Cache, Context, DataPack, Function, TextFileBase
from beet.core.utils import FileSystemPath, JsonDict, extra_field
from pydantic import BaseModel
from tokenstream import InvalidSyntax, TokenStream
from tokenstream.location import set_location

from .ast import AstNode, AstRoot
from .config import CommandTree
from .database import CompilationDatabase, CompilationUnit
from .diagnostic import Diagnostic, DiagnosticCollection, DiagnosticError
from .dispatch import Dispatcher, MutatingReducer, Reducer
from .parse import delegate, get_parsers
from .serialize import Serializer
from .spec import CommandSpec
from .utils import VersionNumber

AstNodeType = TypeVar("AstNodeType", bound=AstNode)
TextFileType = TypeVar("TextFileType", bound=TextFileBase[Any])


logger = logging.getLogger("mecha")


class MechaOptions(BaseModel):
    """Mecha options."""

    version: VersionNumber = "1.17"
    multiline: bool = False
    match: Optional[List[str]] = None
    rules: Dict[str, Literal["ignore", "info", "warn", "error"]] = {}


@dataclass
class Mecha:
    """Class exposing the command api."""

    ctx: InitVar[Optional[Context]] = None
    version: InitVar[VersionNumber] = "1.17"
    multiline: InitVar[bool] = False
    match: Optional[List[str]] = None

    directory: Path = extra_field(init=False)
    cache: Optional[Cache] = extra_field(default=None)

    spec: CommandSpec = extra_field(default=None)

    lint: Dispatcher[AstRoot] = extra_field(init=False)
    normalize: Dispatcher[AstRoot] = extra_field(init=False)
    transform: Dispatcher[AstRoot] = extra_field(init=False)
    optimize: Dispatcher[AstRoot] = extra_field(init=False)
    check: Dispatcher[AstRoot] = extra_field(init=False)

    steps: List[Dispatcher[AstRoot]] = extra_field(default_factory=list)

    serialize: Dispatcher[str] = extra_field(init=False)

    database: CompilationDatabase = extra_field(default_factory=CompilationDatabase)
    diagnostics: DiagnosticCollection = extra_field(
        default_factory=DiagnosticCollection
    )

    def __post_init__(
        self,
        ctx: Optional[Context],
        version: VersionNumber,
        multiline: bool,
    ):
        if ctx:
            opts = ctx.validate("mecha", MechaOptions)
            version = opts.version
            multiline = opts.multiline

            if opts.match is not None:
                self.match = opts.match

            self.directory = ctx.directory

            self.lint = Reducer(levels=opts.rules)
            self.normalize = MutatingReducer(levels=opts.rules)
            self.transform = MutatingReducer(levels=opts.rules)
            self.optimize = MutatingReducer(levels=opts.rules)
            self.check = Reducer(levels=opts.rules)

            if not self.cache:
                self.cache = ctx.cache["mecha"]

        else:
            self.directory = Path.cwd()

            self.lint = Reducer()
            self.normalize = MutatingReducer()
            self.transform = MutatingReducer()
            self.optimize = MutatingReducer()
            self.check = Reducer()

        self.steps.append(self.lint)
        self.steps.append(self.normalize)
        self.steps.append(self.transform)
        self.steps.append(self.optimize)
        self.steps.append(self.check)

        if not self.spec:
            self.spec = CommandSpec(
                multiline=multiline,
                tree=CommandTree.load_from(version=version),
                parsers=get_parsers(version),
            )

        self.serialize = Serializer(self.spec)

    @contextmanager
    def prepare_token_stream(
        self,
        stream: TokenStream,
        multiline: Optional[bool] = None,
    ) -> Iterator[TokenStream]:
        """Prepare the token stream for parsing."""
        with stream.reset(*stream.data), stream.provide(
            spec=self.spec,
            multiline=self.spec.multiline if multiline is None else multiline,
        ):
            with stream.reset_syntax(comment=r"#.*$", literal=r"\S+"):
                with stream.indent(skip=["comment"]), stream.ignore("indent", "dedent"):
                    with stream.intercept("newline", "eof"):
                        yield stream

    @overload
    def parse(
        self,
        source: Union[TextFileBase[Any], List[str], str],
        *,
        filename: Optional[FileSystemPath] = None,
        resource_location: Optional[str] = None,
        multiline: Optional[bool] = None,
        provide: Optional[JsonDict] = None,
    ) -> AstRoot:
        ...

    @overload
    def parse(
        self,
        source: Union[TextFileBase[Any], List[str], str],
        *,
        type: Type[AstNodeType],
        filename: Optional[FileSystemPath] = None,
        resource_location: Optional[str] = None,
        multiline: Optional[bool] = None,
        provide: Optional[JsonDict] = None,
    ) -> AstNodeType:
        ...

    @overload
    def parse(
        self,
        source: Union[TextFileBase[Any], List[str], str],
        *,
        using: str,
        filename: Optional[FileSystemPath] = None,
        resource_location: Optional[str] = None,
        multiline: Optional[bool] = None,
        provide: Optional[JsonDict] = None,
    ) -> Any:
        ...

    def parse(
        self,
        source: Union[TextFileBase[Any], List[str], str],
        *,
        type: Optional[Type[AstNode]] = None,
        using: Optional[str] = None,
        filename: Optional[FileSystemPath] = None,
        resource_location: Optional[str] = None,
        multiline: Optional[bool] = None,
        provide: Optional[JsonDict] = None,
    ) -> Any:
        """Parse the given source into an AST."""
        if using:
            parser = using
        else:
            if not type:
                type = AstRoot
            if not type.parser:
                raise TypeError(f"No parser directly associated with {type}.")
            parser = type.parser

        if isinstance(source, (list, str)):
            source = Function(source)

        if not filename and source.source_path:
            filename = os.path.relpath(source.source_path, self.directory)

        cache_miss = None

        if self.cache and filename:
            ast_path = self.cache.get_path(f"{filename}-ast")

            if not self.cache.has_changed(self.directory / filename):
                try:
                    with ast_path.open("rb") as f:
                        logger.debug("Using cached ast for file %r.", filename)
                        return pickle.load(f)
                except FileNotFoundError:
                    pass

            logger.debug("Updating ast for file %r.", filename)
            cache_miss = ast_path

        stream = TokenStream(source.text)

        try:
            with self.prepare_token_stream(stream, multiline=multiline):
                with stream.provide(**provide or {}):
                    ast = delegate(parser, stream)
        except InvalidSyntax as exc:
            d = Diagnostic(
                level="error",
                message=str(exc),
                hint=resource_location,
                filename=str(filename) if filename else None,
            )
            raise DiagnosticError(DiagnosticCollection([set_location(d, exc)])) from exc
        else:
            if self.cache and filename and cache_miss:
                with cache_miss.open("wb") as f:
                    pickle.dump(ast, f)
            return ast

    @overload
    def compile(
        self,
        source: DataPack,
        *,
        match: Optional[List[str]] = None,
        multiline: Optional[bool] = None,
        report: Optional[DiagnosticCollection] = None,
    ) -> DataPack:
        ...

    @overload
    def compile(
        self,
        source: TextFileType,
        *,
        filename: Optional[FileSystemPath] = None,
        resource_location: Optional[str] = None,
        multiline: Optional[bool] = None,
        report: Optional[DiagnosticCollection] = None,
    ) -> TextFileType:
        ...

    @overload
    def compile(
        self,
        source: Union[List[str], str, AstRoot],
        *,
        filename: Optional[FileSystemPath] = None,
        resource_location: Optional[str] = None,
        multiline: Optional[bool] = None,
        report: Optional[DiagnosticCollection] = None,
    ) -> Function:
        ...

    def compile(
        self,
        source: Union[DataPack, TextFileBase[Any], List[str], str, AstRoot],
        *,
        match: Optional[List[str]] = None,
        filename: Optional[FileSystemPath] = None,
        resource_location: Optional[str] = None,
        multiline: Optional[bool] = None,
        report: Optional[DiagnosticCollection] = None,
    ) -> Union[DataPack, TextFileBase[Any]]:
        """Apply all compilation steps."""
        functions: List[TextFileBase[Any]] = []

        if isinstance(source, DataPack):
            result = source

            if match is None:
                match = self.match

            for key in source.functions.match(*match or ["*"]):
                value = source.functions[key]
                functions.append(value)
                self.database[value] = CompilationUnit(
                    resource_location=key,
                    filename=(
                        os.path.relpath(value.source_path, self.directory)
                        if value.source_path
                        else None
                    ),
                )
        else:
            if isinstance(source, (list, str)):
                source = Function(source)

            if isinstance(source, TextFileBase):
                result = source
                if not filename and source.source_path:
                    filename = os.path.relpath(source.source_path, self.directory)
            else:
                result = Function()

            functions.append(result)
            self.database[result] = CompilationUnit(
                ast=source if isinstance(source, AstRoot) else None,
                resource_location=resource_location,
                filename=str(filename) if filename else None,
            )

        for function in functions:
            compilation_unit = self.database[function]

            if compilation_unit.ast:
                continue

            try:
                compilation_unit.source = function.text
                compilation_unit.ast = self.parse(
                    function,
                    filename=compilation_unit.filename,
                    resource_location=compilation_unit.resource_location,
                    multiline=multiline,
                )
            except DiagnosticError as exc:
                compilation_unit.diagnostics.extend(exc.diagnostics)

        for compilation_step in self.steps:
            for function in functions:
                compilation_unit = self.database[function]

                if not compilation_unit.ast:
                    continue

                with compilation_step.use_diagnostics(compilation_unit.diagnostics):
                    compilation_unit.ast = compilation_step(compilation_unit.ast)

        for function in functions:
            compilation_unit = self.database[function]

            if not compilation_unit.ast:
                continue

            with self.serialize.use_diagnostics(compilation_unit.diagnostics):
                function.text = self.serialize(compilation_unit.ast)

        diagnostics = DiagnosticCollection(
            [
                exc
                for function in functions
                for exc in self.database[function].diagnostics.exceptions
            ]
        )

        if report:
            report.extend(diagnostics)
        else:
            if errors := list(diagnostics.get_all_errors()):
                raise DiagnosticError(DiagnosticCollection(errors))

        return result



__all__ = [
    "CompilationDatabase",
    "CompilationUnit",
]


from dataclasses import dataclass
from typing import Any, Dict, Optional

from beet import Container, TextFileBase
from beet.core.utils import extra_field

from .ast import AstRoot
from .diagnostic import DiagnosticCollection


@dataclass
class CompilationUnit:
    """Information associated with the compilation of a specific function file."""

    ast: Optional[AstRoot] = None
    source: Optional[str] = None
    filename: Optional[str] = None
    resource_location: Optional[str] = None

    diagnostics: DiagnosticCollection = extra_field(init=False)

    def __post_init__(self):
        self.diagnostics = DiagnosticCollection(
            filename=self.filename,
            hint=self.resource_location,
        )


class CompilationDatabase(Container[TextFileBase[Any], CompilationUnit]):
    """Compilation database."""

    index: Dict[str, TextFileBase[Any]]

    def __init__(self):
        super().__init__()
        self.index = {}

    def process(
        self,
        key: TextFileBase[Any],
        value: CompilationUnit,
    ) -> CompilationUnit:
        for index in [value.filename, value.resource_location]:
            if index:
                self.index[index] = key
        return value

    def __delitem__(self, key: TextFileBase[Any]):
        for index in [self[key].filename, self[key].resource_location]:
            if index:
                del self.index[index]
        super().__delitem__(key)



__all__ = [
    "QuoteHelper",
    "QuoteHelperWithUnicode",
    "string_to_number",
    "VersionNumber",
    "split_version",
]


import re
from dataclasses import dataclass, field
from typing import Dict, Tuple, Union

from beet.core.utils import normalize_string
from tokenstream import InvalidSyntax, Token, set_location

from .error import InvalidEscapeSequence

ESCAPE_REGEX = re.compile(r"\\.")
UNICODE_ESCAPE_REGEX = re.compile(r"\\(?:u([0-9a-fA-F]{4})|.)")
AVOID_QUOTES_REGEX = re.compile(r"^[0-9A-Za-z_\.\+\-]+$")


VersionNumber = Union[str, int, float, Tuple[Union[str, int], ...]]


def string_to_number(string: str) -> Union[int, float]:
    """Helper for converting numbers to string and keeping their original type."""
    return float(string) if "." in string else int(string)


def split_version(version: VersionNumber) -> Tuple[int, ...]:
    """Break version number into a tuple of integers."""
    if isinstance(version, (int, float)):
        version = str(version)
    if isinstance(version, str):
        version = tuple(normalize_string(version).split("_"))
    return tuple(map(int, version))


@dataclass
class QuoteHelper:
    """Helper for removing quotes and interpreting escape sequences."""

    escape_regex: "re.Pattern[str]" = ESCAPE_REGEX
    avoid_quotes_regex: "re.Pattern[str]" = AVOID_QUOTES_REGEX

    escape_sequences: Dict[str, str] = field(default_factory=dict)
    escape_characters: Dict[str, str] = field(init=False)

    def __post_init__(self):
        self.escape_characters = {v: k for k, v in self.escape_sequences.items()}

    def unquote_string(self, token: Token) -> str:
        """Remove quotes and substitute escaped characters."""
        if not token.value.startswith(('"', "'")):
            return token.value
        try:
            return self.escape_regex.sub(
                lambda match: self.handle_substitution(token, match),
                token.value[1:-1],
            )
        except InvalidEscapeSequence as exc:
            location = token.location.with_horizontal_offset(exc.index + 1)
            end_location = location.with_horizontal_offset(len(exc.characters))
            exc = InvalidSyntax(f"Invalid escape sequence {exc.characters!r}.")
            raise set_location(exc, location, end_location)

    def handle_substitution(self, token: Token, match: "re.Match[str]") -> str:
        """Handle escaped character sequence."""
        characters = match[0]

        if characters == "\\" + token.value[0]:
            return token.value[0]

        try:
            return self.escape_sequences[characters]
        except KeyError:
            raise InvalidEscapeSequence(characters, match.pos) from None

    def quote_string(self, value: str, quote: str = '"') -> str:
        """Wrap the string in quotes if it can't be represented unquoted."""
        if self.avoid_quotes_regex.match(value):
            return value
        for match, seq in self.escape_characters.items():
            value = value.replace(match, seq)
        return quote + value.replace(quote, "\\" + quote) + quote


@dataclass
class QuoteHelperWithUnicode(QuoteHelper):
    """Quote helper that handles unicode escape sequences."""

    escape_regex: "re.Pattern[str]" = UNICODE_ESCAPE_REGEX

    def handle_substitution(self, token: Token, match: "re.Match[str]") -> str:
        if unicode_hex := match[1]:
            return chr(int(unicode_hex, 16))
        return super().handle_substitution(token, match)



__all__ = [
    "BasicLinter",
]


from beet import Context
from tokenstream import set_location

from mecha import (
    AstCommand,
    AstSelector,
    Diagnostic,
    DiagnosticCollection,
    Mecha,
    Reducer,
    rule,
)


def beet_default(ctx: Context):
    mc = ctx.inject(Mecha)
    mc.lint.extend(BasicLinter())


class BasicLinter(Reducer):
    """Linter with basic rules."""

    @rule(AstCommand, identifier="execute:subcommand")
    def execute_run(self, node: AstCommand):
        if isinstance(clause := node.arguments[0], AstCommand):
            if clause.identifier == "execute:run:subcommand":
                d = Diagnostic("warn", "Redundant `execute run` clause.")
                raise set_location(d, node, clause.arguments[0].location)

    @rule(AstCommand, identifier="execute:run:subcommand")
    def run_execute(self, node: AstCommand):
        if isinstance(clause := node.arguments[0], AstCommand):
            if clause.identifier == "execute:subcommand":
                d = Diagnostic("warn", "Redundant `run execute` clause.")
                raise set_location(d, node, clause.arguments[0].location)

    @rule(AstSelector)
    def selector_argument_order(self, node: AstSelector):
        order = [
            "type",
            "gamemode",
            "inverted gamemode",
            "team",
            "inverted team",
            "inverted type",
            "tag",
            "inverted tag",
            "name",
            "inverted name",
            "scores",
            "predicate",
            "inverted predicate",
            "advancements",
            "nbt",
        ]
        conflict = [-1] * len(order)

        with DiagnosticCollection() as diagnostics:
            for i, arg in enumerate(node.arguments):
                name = "inverted " * arg.inverted + arg.key.value

                try:
                    index = order.index(name)
                except ValueError:
                    continue

                j = conflict[index]

                if j >= 0:
                    bad_arg = node.arguments[j]
                    bad_arg_name = "inverted " * bad_arg.inverted + bad_arg.key.value

                    d = Diagnostic(
                        level="warn",
                        message=f"{name.capitalize()} argument should go before {bad_arg_name}.",
                    )

                    bad_arg_message = f"Specifying the {bad_arg_name} argument before {name} leads to unnecessary processing."
                    d.labels[bad_arg] = bad_arg_message

                    diagnostics.add(set_location(d, arg))

                for conflict_index in range(index):
                    if conflict[conflict_index] < 0:
                        conflict[conflict_index] = i
